﻿//=========================================================================
// Copyright 2009 Sergey Antonov <sergant_@mail.ru>
// 
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation.
// 
// See the file COPYING.TXT for the full text of the license, or see
// http://www.gnu.org/licenses/gpl-2.0.txt
// 
//=========================================================================

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Win32;
using System.Linq;
using System.IO;

namespace HgSccHelper
{
	static public class Util
	{
		//-----------------------------------------------------------------------------
		public static string Quote(this string str)
		{
			if (str.Length == 0)
				return str;

			if (str.Length >= 2)
			{
				if (str[0] == '\"' && str[str.Length - 1] == '\"')
					return str;
			}

			return "\"" + str + "\"";
		}

		//-----------------------------------------------------------------------------
		public static string UnQuote(this string str)
		{
			if (str.Length >= 2)
			{
				if (str[0] == '\"' && str[str.Length - 1] == '\"')
					return str.Substring(1, str.Length - 2);
			}

			return str;
		}

		//-----------------------------------------------------------------------------
		public static string EscapeQuotes(this string str)
		{
			return str.Replace("\"", "\\\"");
		}

		//-----------------------------------------------------------------------------
		public static string UnEscapeQuotes(this string str)
		{
			return str.Replace("\\\"", "\"");
		}

		//------------------------------------------------------------------
		public static string ShortSHA1(this string str)
		{
			return str.Substring(0, 12);
		}

		//------------------------------------------------------------------
		public static string GetTempFileNameForFile(string file)
		{
			string temp_folder = Path.GetTempPath();
			string temp_name = Path.GetRandomFileName() + "_" + Path.GetFileName(file);
			string temp_file = Path.Combine(temp_folder, temp_name);

			return temp_file;
		}

		//------------------------------------------------------------------
		public static bool FindOrAdd<TKey, TValue>(this Dictionary<TKey, TValue> dictionary,
			TKey key, ref TValue v)
		{
			TValue found;
			if (dictionary.TryGetValue(key, out found))
			{
				v = found;
				return true;
			}

			dictionary[key] = v;
			return false;
		}

		//------------------------------------------------------------------
		public static T[] ToArray<T>(this HashSet<T> hash_set)
		{
			var array = new T[hash_set.Count];
			hash_set.CopyTo(array);
			return array;
		}

		//-----------------------------------------------------------------------------
		public static bool GetRelativePath(string working_dir, string path, out string relative)
		{
			relative = string.Empty;
			string f = path.ToLower();
			if (!f.StartsWith(working_dir.ToLower()))
				return false;

			if (path.Length == working_dir.Length)
				relative = "";
			else
				relative = path.Substring(working_dir.Length + 1);
			return true;
		}

		//-----------------------------------------------------------------------------
		static public string LookupRegistry(string reg_key_name, string val_name)
		{
			var reg_key = Registry.CurrentUser.OpenSubKey(reg_key_name, RegistryKeyPermissionCheck.ReadSubTree);
			if (reg_key == null)
				reg_key = Registry.LocalMachine.OpenSubKey(reg_key_name, RegistryKeyPermissionCheck.ReadSubTree);

			if (reg_key != null)
			{
				var val = reg_key.GetValue(val_name);
				if (val != null)
					return val.ToString();
			}

			return "";
		}

		//-----------------------------------------------------------------------------
		static public string FindExisting(string pathcommand)
		{
			return FindExisting(pathcommand, new[] { "" });
		}

		//-----------------------------------------------------------------------------
		static public string FindExisting(string pathcommand, string[] pathexts)
		{
			foreach (var ext in pathexts)
			{
				var exectutable = pathcommand + ext;
				if (System.IO.File.Exists(exectutable))
					return exectutable;
			}

			return "";
		}

		//-----------------------------------------------------------------------------
		static public string FindExe(string command)
		{
			var pathext = Environment.GetEnvironmentVariable("PATHEXT") ?? ".COM;.EXE;.BAT;.CMD";
			var path_sep = new[] { ';' };
			var pathexts = pathext.ToLower().Split(path_sep, StringSplitOptions.RemoveEmptyEntries);

			var command_ext = System.IO.Path.GetExtension(command).ToLower();
			if (pathexts.Contains(command_ext))
				pathexts = new string[] { "" };

			if (command.Contains('/') || command.Contains('\\'))
				return FindExisting(command, pathexts);

			var path_env = Environment.GetEnvironmentVariable("PATH") ?? "";
			foreach (var path in path_env.Split(path_sep, StringSplitOptions.RemoveEmptyEntries))
			{
				var executable = FindExisting(System.IO.Path.Combine(path.UnQuote(), command), pathexts);
				if (!String.IsNullOrEmpty(executable))
					return executable;
			}

			return FindExisting(Environment.ExpandEnvironmentVariables(command), pathexts);
		}
	}
}